/* *           Copyright (c) 2004, Daniel M. Bikel.
 *                         All rights reserved.
 * 
 *                Developed at the University of Pennsylvania
 *                Institute for Research in Cognitive Science
 *                    3401 Walnut Street
 *                    Philadelphia, Pennsylvania 19104
 * 			
 * 
 * For research or educational purposes only.  Do not redistribute.  For
 * complete license details, please read the file LICENSE that accompanied
 * this software.
 * 
 * DISCLAIMER
 * 
 * Daniel M. Bikel makes no representations or warranties about the suitability of
 * the Software, either express or implied, including but not limited to the
 * implied warranties of merchantability, fitness for a particular purpose, or
 * non-infringement. Daniel M. Bikel shall not be liable for any damages suffered
 * by Licensee as a result of using, modifying or distributing the Software or its
 * derivatives.
 * 
 */
    package danbikel.parser.arabic;

import danbikel.parser.Constants;
import danbikel.parser.HeadFinder;
import danbikel.parser.Language;
import danbikel.parser.Treebank;
import danbikel.parser.Settings;
import danbikel.parser.Nonterminal;
import danbikel.parser.Word;
import danbikel.parser.Words;
import danbikel.util.*;
import danbikel.lisp.*;
import java.util.*;
import java.io.*;

/**
 * Provides methods for language-specific processing of training parse trees.
 * Even though this subclass of {@link danbikel.parser.Training} is
 * in the default English language package, its primary purpose is simply
 * to fill in the {@link #argContexts}, {@link #semTagArgStopSet} and
 * {@link #nodesToPrune} data members using a metadata resource.  If this
 * capability is desired in another language package, this class may be
 * subclassed.
 * <p>
 * This class also re-defined the method
 * {@link danbikel.parser.lang.AbstractTraining#hasPossessiveChild(Sexp)}.
 */
public class Training2NoBaseNPs extends danbikel.parser.arabic.Training2 {

    /* must be a better way to do this */
  private final static Symbol tagMapSym = Symbol.add("tag-map");

  public Training2NoBaseNPs() throws FileNotFoundException, IOException {
      super();
      System.err.println("in arabic Training 2NoBaseNPs");
  }

  public Sexp preProcess(Sexp tree) {
    transformTags(tree);
    prune(tree);
    // addBaseNPs(tree);
    //repairBaseNPs(tree);
    //addGapInformation(tree);
    //relabelSubjectlessSentences(tree);
    removeNullElements(tree);
    raisePunctuation(tree);
    identifyArguments(tree);
    stripAugmentations(tree);
    return tree;
  }


  protected void readMetadataHook(Symbol dataType,
				  int metadataLen, SexpList metadata) {
    if (dataType == tagMapSym) {
      for (int i = 1; i < metadataLen; i++) {
	SexpList mapping = metadata.get(i).list();
	TagMap2.add(mapping.symbolAt(0), mapping.symbolAt(1));
      }
    }
  }

  public SexpList preProcessTest(SexpList sentence,
				 SexpList originalWords, SexpList tags) {
    if (tags == null)
      return super.preProcessTest(sentence, originalWords, tags);
    SexpList processed = new SexpList(2);
    processed.add(sentence);
    int numWords = sentence.size();
    for (int i = 0; i < numWords; i++) {
      Symbol origWord = originalWords.symbolAt(i);

      SexpList currWordTags = tags.listAt(i);
      int numTags = currWordTags.size();

      for (int tagIdx = 0; tagIdx < numTags; tagIdx++) {

	Symbol origTag = currWordTags.symbolAt(tagIdx);
	Word word = danbikel.parser.Words.get(origWord, origTag);

	currWordTags.set(tagIdx, TagMap2.transformTag(word));
      }

      tags.set(i, currWordTags);
    }
    processed.add(tags);
    System.out.println("preprocesstest Training2 after tags = " + tags);
    return processed;
  }

  protected Sexp transformTags(Sexp tree) {
    if (Language.treebank().isPreterminal(tree) &&
	!Language.treebank().isNullElementPreterminal(tree)) {
      Word word = Language.treebank().makeWord(tree);
      Symbol newTag = TagMap2.transformTag(word);
      tree.list().set(0, newTag);
    }
    else if (tree.isList()) {
      SexpList treeList = tree.list();
      int treeListLen = treeList.length();
      for (int i = 1; i < treeListLen; i++)
	transformTags(treeList.get(i));
    }
    return tree;
  }


    /* only strip default arg, not the new args like SBJ
     */
  protected void stripArgAugAndCanonicalizeBases(Sexp tree) {
      if (treebank.isPreterminal(tree)) {
	  /*
	    SexpList treeList = tree.list();
	    Symbol currLabel = treeList.symbolAt(0);
	    treeList.set(0, treebank.getCanonical(currLabel));
	  */
	  return;
      }
      else if (tree.isList()) {
	  SexpList treeList = tree.list();
	  Symbol currLabel = treeList.symbolAt(0);
	  treeList.set(0, stripArgAugAndCanonicalizeBase(currLabel));
	  int treeListLen = treeList.length();
	  for (int i = 1; i < treeListLen; i++)
	      stripArgAugAndCanonicalizeBases(treeList.get(i));
      }
  }
    public Symbol stripArgAugAndCanonicalizeBase(Symbol label) {
	//    Symbol canonicalArgLabel = (Symbol)canonicalArgCache.get(label);
	//    if (canonicalArgLabel == null) {
	Symbol newLabel=removeDefaultArgAugmentation(label,new Nonterminal());
	Nonterminal parsedLabel =
	    treebank.parseNonterminal(newLabel);
	parsedLabel.base = treebank.getCanonical(parsedLabel.base);
	return parsedLabel.toSymbol();
    }
    //      if (argAugmentations.contains(augmentations.get(i))) {
  protected Symbol removeDefaultArgAugmentation(Symbol label,
                                         Nonterminal nonterminal) {
    Nonterminal parsedLabel = treebank.parseNonterminal(label, nonterminal);
    boolean removedSomething = false;
    SexpList augmentations = parsedLabel.augmentations;
    for (int i = 0; i < augmentations.length(); i++) {
	if (defaultArgAugmentation==augmentations.get(i)) {
	    i--; // move index to delimeter that preceded this arg augmentation
	    augmentations.remove(i);  // remove delimeter
	    augmentations.remove(i);  // remove arg augmentation
	    removedSomething = true;
	    i--; // decrement index again to offset increment in for loop
	}
    }
    return removedSomething ? parsedLabel.toSymbol() : label;
  }

  public void postProcess(Sexp tree) {
    //unrepairBaseNPs(tree);
      //    super.postProcess(tree);
      System.out.println("before remove basenps = " + tree);
    removeOnlyChildBaseNPs(tree);
       System.out.println("after remove basenps = " + tree);
    //    canonicalizeNonterminals(tree);
    stripArgAugAndCanonicalizeBases(tree);
  }
}
