/* *           Copyright (c) 2004, Daniel M. Bikel.
 *                         All rights reserved.
 * 
 *                Developed at the University of Pennsylvania
 *                Institute for Research in Cognitive Science
 *                    3401 Walnut Street
 *                    Philadelphia, Pennsylvania 19104
 * 			
 * 
 * For research or educational purposes only.  Do not redistribute.  For
 * complete license details, please read the file LICENSE that accompanied
 * this software.
 * 
 * DISCLAIMER
 * 
 * Daniel M. Bikel makes no representations or warranties about the suitability of
 * the Software, either express or implied, including but not limited to the
 * implied warranties of merchantability, fitness for a particular purpose, or
 * non-infringement. Daniel M. Bikel shall not be liable for any damages suffered
 * by Licensee as a result of using, modifying or distributing the Software or its
 * derivatives.
 * 
 */
    package danbikel.parser.ppceme;

import java.util.BitSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.StringTokenizer;
import danbikel.lisp.*;
import danbikel.parser.Language;
import danbikel.parser.Nonterminal;

/*
 * Provides data and methods speciifc to the structures found in the English
 * Treebank (the
 * <a target="_blank"
 * href="http://www.cis.upenn.edu/~treebank/">Penn Treebank</a>) or any other
 * treebank that conforms to the Treebank II annotation guidelines for
 * <a target="_blank" href="ftp://ftp.cis.upenn.edu/pub/treebank/doc/tagguide.ps.gz">part-of-speech tagging</a>
 * and
 * <a target="_blank"
 * href="ftp://ftp.cis.upenn.edu/pub/treebank/doc/manual/">bracketing</a>.
 */

/**
 * Provides data and methods speciifc to the structures found in the English
 * Treebank (the Penn Treebank) or any other treebank that conforms to the
 * Treebank II annotation guidelines for part-of-speech tagging and bracketing.
 */
public class Treebank extends danbikel.parser.lang.AbstractTreebank2 {

  static final Symbol IP = Symbol.add("IP");

  public boolean isIP(Symbol label) {
    return getCanonical(label) == IP;
  }

  // the characters that are delimiters of augmented nonterminal labels
  private final static String augmentationDelimStr = "-=+";

  // basic nodes in the English Treebank that will be transformed in a
  // preprocessing phase
  static final Symbol NP = Symbol.add("NP");
    /* BS:
      sentences are tagged IP (with various subcategories, such IP-MAT,  
      IP-SUB, IP-PPL, etc.)
    */
  static final Symbol S = Symbol.add("IP");
  // transformed nodes
  static final Symbol baseNP = Symbol.add(NP + "B");
  static final Symbol baseNPArg = Symbol.add(baseNP + "-A");

    /* BS:
       here i'm guessing...  not all IP-PPL's are subjectless, but mostly  
       they are, and pretty much all subjectless clauses will be IP-PPL
       note that our subjectless sentences don't have null subjects (the  
       understood subject is simply missing), and as usual, there is no vp
    */
  static final Symbol subjectlessS = Symbol.add(S + "G");
  static final Symbol subjectAugmentation = Symbol.add("SBJ");

  // other basic nodes
  static final Symbol WHNP = Symbol.add("WHNP");
  static final Symbol CC = Symbol.add("CONJ");
  static final Symbol comma = Symbol.add(",");

    /* BS:
      i'm guessing that lrb, rrb refer to left/right round bracket.  not  
      exactly sure what l/rcb stand for.
      in any event, there are some instances of <paren> and <bracket> in  
      the corpus.  the close counterparts are </paren> and </bracket>, and
      potentially <$$paren> and <$$bracket>
    */

    /* not right */
  static final Symbol openParen = Symbol.add("OPEN");
  static final Symbol closeParen = Symbol.add("CLOSE");

    /* BS:we use *T* instead of NULL for traces */
    /* TODO: Actually it's also just *.  This may not be
       relevant, though - see isNullElementPreterminal
    */
     
  static final Symbol nullElementPreterminal = Symbol.add("-NONE-");

  // possessive part of speech
  static final Symbol possessivePos = Symbol.add("POS");

  // exception nonterminals for defaultParseNonterminal
  static final Symbol[] nonterminalExceptionArr = {};

  private static Symbol[][] canonicalLabelMapData = {
    {baseNP, NP},
    {subjectlessS, S},
  };
  private static HashMap canonicalLabelMap =
    new HashMap(canonicalLabelMapData.length);
  static {
    for (int i = 0; i < canonicalLabelMapData.length; i++)
      canonicalLabelMap.put(canonicalLabelMapData[i][0],
                            canonicalLabelMapData[i][1]);
  }

    /* BS:
       the only sentence-internal punctuation tag should be comma (there  
       might be erroneous periods, however).
    */

    /* probably don't need : here
       I guess period isn't here because it gets pruned so is in
       training-metadata.lisp.
       But Dan had put period in puncToRaiseElements for Portugese.
       Seems inconsistent
    */
  private static String[] puncToRaiseElements = {",", ":"};

  private static Set puncToRaise = new HashSet();
  static {
    for (int i = 0; i < puncToRaiseElements.length; i++)
      puncToRaise.add(Symbol.add(puncToRaiseElements[i]));
  }

    /* BS:
       if the BE, DO, and HV series can profitably be reduced to VB, that's  
       fine
       note that DAN, HAN, VAN are passive participles and perhaps do not  
       belong here, as they won't license objects.
    */

    private static String [] verbTagStrings = 
    {"BAG", "BE", "BED", "BEI",  
     "BEN", "BEP",
     "DAG", "DAN", "DO", "DOD", "DOI", "DON", "DOP",
     "HAG", "HAN", "HV", "HVD", "HVI", "HVN", "HVP",
     "VAG", "VAN", "VB", "VBD", "VBI", "VBN", "VBP"};

    private static Set verbTags = new HashSet();
  static {
    for (int i = 0; i < verbTagStrings.length; i++)
      verbTags.add(Symbol.add(verbTagStrings[i]));
  }

  /**
   * Constructs an English <code>Treebank</code> object.
   */
  public Treebank() {
    super();
    nonterminalExceptionSet = nonterminalExceptionArr;
  }

  /**
   * Returns <code>true</code> if <code>tree</code> represents a preterminal
   * subtree (part-of-speech tag and word).  Specifically, this method
   * returns <code>true</code> if <code>tree</code> is an instance of
   * <code>SexpList</code>, has a length of 2 and has a first list element
   * of type <code>Symbol</code>.
   */
  public final boolean isPreterminal(Sexp tree) {
    return (tree.isList() &&
	    tree.list().length() == 2 &&
	    tree.list().get(0).isSymbol() &&
	    tree.list().get(1).isSymbol());
  }

  /**
   * Returns <code>true</code> is the specified nonterminal label represents a
   * sentence in the Penn Treebank, that is, if the canonical version of
   * <code>label</code> is equal to <code>&quot;S&quot;</code>.
   * @see Training#relabelSubjectlessSentences(Sexp)
   */
  public boolean isSentence(Symbol label) { return getCanonical(label) == S; }

  public Symbol sentenceLabel() { return S; }

  /**
   * Returns the symbol that
   * {@link Training#relabelSubjectlessSentences(Sexp)}
   * will use for sentences that have no subjects.
   */
  public Symbol subjectlessSentenceLabel() { return subjectlessS; }

  public Symbol subjectAugmentation() { return subjectAugmentation; }

  /**
   * Returns <code>true</code> if the specified S-expression represents a
   * preterminal whose terminal element is the null element
   * (<code>&quot;-NONE-&quot;</code>) for the Penn Treebank.
   * @see Training#relabelSubjectlessSentences(Sexp)
   */

  public boolean isNullElementPreterminal(Sexp tree) {

	return (isPreterminal(tree) &&
	tree.list().get(0).symbol() == nullElementPreterminal);
	/*
      return (isPreterminal(tree) &&
	      tree.list().get(1).symbol().toString().charAt(0) == '*');
	*/
  }
    

  /**
   * Returns <code>true</code> if the specified S-expression is a preterminal
   * whose part of speech is <code>&quot;,&quot;</code> or
   * <code>&quot;:&quot;</code>.
   */
  public boolean isPuncToRaise(Sexp preterm) {
    return (isPreterminal(preterm) &&
	    puncToRaise.contains(preterm.list().first()));
  }

  public boolean isPunctuation(Symbol tag) {
      /* which should it be? */

    //return puncToRaise.contains(stripAugmentation(tag));
    return puncToRaise.contains(tag);
  }

  /**
   * Returns <code>true</code> if the specified S-expression represents
   * a preterminal that is the possessive part of speech.  This method is
   * intended to be used by implementations of {@link
   * Training#addBaseNPs(Sexp)}.
   */

  public boolean isPossessivePreterminal(Sexp tree) {
    return (isPreterminal(tree) &&
	    tree.list().get(0).symbol() == possessivePos);
  }

  /**
   * Returns <code>true</code> if the canonical version of the specified label
   * is an NP for for English Treebank.
   *
   * @see Training#addBaseNPs(Sexp)
   */
  public boolean isNP(Symbol label) {
    return getCanonical(label) == NP;
  }

  public boolean isBaseNP(Symbol label) {
    return label == baseNP || label == baseNPArg;
  }

  /**
   * Returns the symbol with which {@link Training#addBaseNPs(Sexp)} will
   * relabel base NPs.
   *
   * @see Training#addBaseNPs
   */
  public Symbol baseNPLabel() { return baseNP; }

  /**
   * Returns <code>true</code> if the canonical version of the specified label
   * is a WHNP in the English Treebank.
   *
   * @see Training#addGapInformation(Sexp)
   */
  public boolean isWHNP(Symbol label) { return getCanonical(label) == WHNP; }

  /**
   * Returns the symbol that {@link Training#addBaseNPs(Sexp)} should
   * add as a parent if a base NP is not dominated by an NP.
   */
  public Symbol NPLabel() { return NP; }

  /**
   * Returns <code>true</code> if <code>label</code> is equal to the symbol
   * whose print name is <code>&quot;CC&quot;</code>.
   */
  public boolean isConjunction(Symbol label) {
    //return getCanonical(label) == CC;
    return label == CC;
  }

  public boolean isConjunctionPhrase(Symbol label) {
    //return getCanonical(label) == CC;
      String str=label.toString();
      return (str.startsWith("CONJP"));
  }


  /**
   * Returns <code>true</code> if <code>preterminal</code> represents a
   * terminal with one of the following parts of speech: <tt>VB, VBD, VBG,
   * VBN, VBP</tt> or <tt>VBZ</tt>.  It is an error to call this method
   * with a <code>Sexp</code> object for which {@link #isPreterminal(Sexp)}
   * returns <code>false</code>.<br>
   *
   * @param preterminal the preterminal to test
   * @return <code>true</code> if <code>preterminal</code> is a verb
   */
  public boolean isVerb(Sexp preterminal) {
    return isVerbTag(preterminal.list().get(0).symbol());
  }

  public boolean isVerbTag(Symbol tag) {
    return verbTags.contains(tag);
  }

  public boolean isComma(Symbol word) {
    return word == comma;
  }

  public boolean isLeftParen(Symbol word) {
    return word == openParen;
  }

  public boolean isRightParen(Symbol word) {
    return word == closeParen;
  }

  /**
   * Returns a canonical mapping for the specified nonterminal label; if
   * <code>label</code> already is in canonical form, it is returned.
   * The canonical mapping refers to transformations performed on nonterminals
   * during the training process.  Before obtaining a label's canonical form,
   * it is also stripped of all Treebank augmentations, meaning that only
   * the characters before the first occurrence of '-', '=' or '|' are kept.
   *
   * @return a <code>Symbol</code> with the same print name as
   * <code>label</code>, except that all training transformations and Treebank
   * augmentations have been undone and stripped
   */
  public final Symbol getCanonical(Symbol label) {
    for (int i = 0; i < nonterminalExceptionSet.length; i++)
      if (label == nonterminalExceptionSet[i])
	return label;
    Symbol strippedLabel = stripAugmentation(label);
    Symbol mapEntry = (Symbol)canonicalLabelMap.get(strippedLabel);
    return ((mapEntry == null) ? strippedLabel : mapEntry);
  }

  public final Symbol getCanonical(Symbol label, boolean stripAugmentations) {
    if (stripAugmentations)
      return getCanonical(label);
    else {
      Symbol mapEntry = (Symbol)canonicalLabelMap.get(label);
      return (mapEntry == null ? label : mapEntry);
    }
  }

  /**
   * Calls {@link Treebank#defaultParseNonterminal(Symbol, Nonterminal)} with
   * the specified arguments.
   * @param label to the nonterminal label to parse
   * @param nonterminal the <code>Nonterminal</code> object to fill with
   * the components of <code>label</code>
   */
  public Nonterminal parseNonterminal(Symbol label, Nonterminal nonterminal) {
    defaultParseNonterminal(label, nonterminal);
    return nonterminal;
  }

  /**
   * Returns a string of the three characters that serve as augmentation
   * delimiters in the Penn Treebank: <code>&quot;-=|&quot;</code>.
   */
  public String augmentationDelimiters() { return augmentationDelimStr; }
}
