package danbikel.parser.constraints;

import danbikel.util.*;
import danbikel.lisp.*;
import danbikel.parser.*;
import java.util.*;
/**
 * An implementation of a constraint to sit in a tree structure of constraints
 * that represent a subgraph (certain brackets) of a tree, constraining a
 * decoder to pursue only theories that contain the brackets of the
 * constraint set of these objects.
 */

public class PartialTreeConstraint2 implements Constraint, SexpConvertible {
    /**
     * Inner class to provide hidden "root" constraint that is always
     * partially satisfied, but never <i>fully satisfied</i> (much like some
     * people I know).
     *
     */
    static class Root extends PartialTreeConstraint2 {
	public Root(PartialTreeConstraint2 observedRoot) {
	    label = Symbol.get("*ROOT*");
	    children = new ArrayList(1);
	    children.add(observedRoot);
	    this.start = observedRoot.start();
	    this.end = observedRoot.end();
	    // modify observedRoot's parent!!!
	    observedRoot.parent = this;
	}
	public boolean isLeaf() { return false; }
	public boolean isViolatedByChild(Item childItem) { return false; }
	public boolean isViolatedBy(Item item) { return false; }
	public boolean isSatisfiedBy(Item item) { return true; }
	public boolean isLocallySatisfiedBy(Item item) { 
	    boolean matchesNegative=matchesNegativeConstraint(item);
	    if (matchesNegative) {
		//		System.out.println("isLocallySatisfied root ruling out " + item);
		return false;
	    } else {
		//		System.out.println("isLocallySatisfied root okay " + item);
		return true;
	    }
	}
    }

    private final static String negativeConstraintsProperty=
	"parser.decoder.constraints.negativeConstraints";

    /**
     * An ArrayList of negative constraints.  Each negative constraint
     * is a Nonterminal.  This is initialized in the static area,
     * since all PartialTreeConstraint2 objects share the same 
     * negative constraints.
     */
   
    protected static ArrayList negativeConstraints=null;
    
    static Treebank treebank=null;
    static Symbol NP=null;
    static Symbol baseNP=null;

    static {
	treebank = Language.treebank();
	NP = treebank.NPLabel();
	baseNP=treebank.baseNPLabel();

	String negativeConstraintsStr=
	    Settings.get(negativeConstraintsProperty);
	System.out.println("negative Constraints= " + negativeConstraintsStr);
	if (negativeConstraintsStr==null) {
	    System.out.println("no negative constraints");
	} else {
	    negativeConstraints=new ArrayList(1);
	    String[] negativeConstraintsStrArray=negativeConstraintsStr.split(",");
	    for (int j=0; j<negativeConstraintsStrArray.length; j++) {
		String singleNegativeConstraintStr=negativeConstraintsStrArray[j];
		System.out.println("singleNegativeConstraintStr="+
				   singleNegativeConstraintStr);
		Symbol constraintSymbol=Symbol.add(singleNegativeConstraintStr);
		Nonterminal constraintNT=new Nonterminal();
		Language.treebank().parseNonterminal(constraintSymbol,constraintNT);
		negativeConstraints.add(constraintNT);
	    }
	    for (int i=0; i<negativeConstraints.size(); i++) {
		System.out.println("negative Constraint #" + i + ") "+ 
				   negativeConstraints.get(i));
	    }
	}
    }

    /**
     * The parent constraint of this constraint.
     */
    protected PartialTreeConstraint2 parent;
    /**
     * All child constraints of this constraint.
     */
    protected List children;
    /**
     * The nonterminal label associated with this constraint.
     */
    protected Symbol label;
    /**
     * A {@link Nonterminal} object for use with
     * {@link Treebank#parseNonterminal(Symbol,Nonterminal)}.
     */
    protected Nonterminal nt = new Nonterminal();
    /**
     * A {@link Nonterminal} object for use with
     * {@link Treebank#parseNonterminal(Symbol,Nonterminal)}.
     */
    protected Nonterminal otherNT3 = new Nonterminal();
    /**
     * A {@link Nonterminal} object for use with
     * {@linkplain #matchesNegativeConstraint}.
     */


    protected Nonterminal otherNT2 = new Nonterminal();
    protected Nonterminal otherNT = new Nonterminal();
    /**
     * The starting word index of the syntactic subtree covered by this
     * constraint.
     */
    protected int start;
    /**
     * The ending word index of the syntactic subtree covered by this constraint.
     */
    protected int end;

    /**
     * Contains whether this constraint has been <i>partially satisfied</i>. A
     * constraint of this type has been partially satisfied if, during the
     * bottom-up decoding process, its label and other local information are
     * found to be consisting with the derivation being pursued.
     */
    //  protected boolean satisfied;
    /**
     * Contains whether this constraint has been <i>fully satisfied</i>.  A
     * constraint of this type is fully satisfied if it is both {@linkplain
     * #satisfied partially satisfied} and if all child constraints have been
     * {@linkplain #fullySatisfied fully satisfied}.
     */
    //  protected boolean fullySatisfied;



    /**
     * Constructs a tree of constraints rooted at the specified syntactic tree.
     *
     * @param tree a possibly-underspecified syntactic tree representing nodes
     * and spans that must be produced by any derivation pursued by the
     * decoder
     */
    public PartialTreeConstraint2(Sexp tree) {
	this(null, tree, new IntCounter(0));
    }

    /**
     * Constructs an empty constraint tree.
     */
    protected PartialTreeConstraint2() {}

    /**
     * A helper constructor for constructing a tree of constraints rooted
     * at the specified subtree with the specified <code>parent</code>.
     *
     * @param parent the parent of the specified tree, or <code>null</code>
     * if the specified tree has no parent
     * @param tree the tree for which to construct an isomorphic tree of
     * constraints
     * @param currWordIdx the start index of the span covered by the specified
     * tree, threaded through the recursive calls to this constructor
     */
    protected PartialTreeConstraint2(PartialTreeConstraint2 parent,
				     Sexp tree, IntCounter currWordIdx) {
	if (Language.treebank().isPreterminal(tree)) {
	    this.parent = parent;
	    children = Collections.EMPTY_LIST;
	    Word word = Language.treebank().makeWord(tree);
	    label = word.tag();
	    Language.treebank().parseNonterminal(label, nt);
	    start = end = currWordIdx.get();
	    currWordIdx.increment();
	}
	else {
	    SexpList treeList = tree.list();
	    int treeListLen = treeList.length();

	    //label = Language.treebank().getCanonical(treeList.symbolAt(0));
	    label = treeList.symbolAt(0);
	    /* what a hack!  don't always want to do this! 
	       For now, if the contraint has function tags, just ignore them,
	       since constraint parsing with function tags is a disaster.
	       But of course this should be a property or something
	     */
	    label=Language.treebank().stripAugmentation(label);
	    Language.treebank().parseNonterminal(label, nt);
	    start = currWordIdx.get();
	    this.parent = parent;
	    children = new ArrayList(treeListLen - 1);
	    for (int i = 1; i < treeListLen; i++) {
		Sexp child = treeList.get(i);
		children.add(new PartialTreeConstraint2(this, child, currWordIdx));
	    }
	    end = currWordIdx.get() - 1;
	}
    }

    /* 
       ==============================================================
       RELEVANT CODE FOR ADDUNARIES
       ==============================================================
    */

    /**
     * not used for PartialTreeConstraint2, but here to satisfy interface
     */
    public Constraint getParent() { 
	return null;
    }

    /**
     * Returns the parent of this constraint if this constraint has been {@link
     * #fullySatisfied}; otherwise, returns this constraint.  If the span and 
     * label are an exact match, we can assume that it was checked by 
     * {@link #isSatisfiedBy}, called by addStopProbs in Decoder.java, and 
     * so it must be fully satisfied.
     * <p/>
     * <b>Implementation note</b>: While this method does not always return the
     * parent constraint of this constraint, the semantics of this method are
     * appropriate for imposing partial tree constraints during a bottom-up
     * decoding process.
     *
     * @return the parent of this constraint if this constraint has been {@link
     *         #fullySatisfied}; otherwise, returns this constraint.
     */

    public Constraint getParent(Item item) {
	if (item==null) {
	    System.out.println("WARNING: attempted getParent on null "+
			       "item, constraint="+this);
	    return this;
	}
	// why did I put this in? 
	if (label==Symbol.get("*ROOT*")) {
	    //	    System.out.println("WARNING: attempted getParent on ROOT"+
	    // " item, constraint="+this);
	    return null;
	}
	// here, unlike with isSatisfiedBy, we do have to worry about
	// item being a preterminal.  We always want to return the
	// parent constraint of a preterminal's constraint, and we
	// don't care if the labels match or not (we don't want to
	// enforce that the POS tags match, which is handled by the
	// useOnlySuppliedTags flag.  An alternative I guess is to
	// make the constraints for pos tags be the wildcard, so
	// no special handling is needed for preterminal contraints.
	// but I didn't do that.
	//	return (spanMatches(item) && subsumes(item)) ? parent : this;
	return matches((CKYItem)item) ? parent : this;
    }
    
    
    /**
     * Returns whether the specified child chart item violates this constraint
     * by having a span beyond the boundaries of this constraint's span.
     * @param childItem the child item to be tested
     * @return whether the specified child chart item violates this constraint
     * by having a span beyond the boundaries of this constraint's span.
     */
    public boolean isViolatedByChild(Item childItem) {
	/*
	if (!spanOK(childItem)) {
	    System.out.println("UNEXPECTED isViolatedByChild !spanOK " + 
			       "\n    "+childItem+
			       "\n    "+this);
	}
	*/
	return !spanOK(childItem);


	/* note: I commented out the following code.  I don't think it's
	   with all the other changes put it
	*/
	/*
	  if (!spanOK(childItem))
	  return true;
	  // if the child item's constraint has a span equal to the child item's span
	  // then the child item's constraint must be fully satisfied
	  PartialTreeConstraint2 ptc =
	  (PartialTreeConstraint2)childItem.getConstraint();
	  if (ptc.spanMatches(childItem))
	  return !ptc.isFullySatisfied(childItem);
	  else
	  return false;
	*/
	}


    /** 
     * @param item the item representing a partial contituent created in
     * {@link Decoder#joinItems}.
     * @return true if the item is inconsistent with the constraint subtree
     * rooted at the current constraint (which is the contraint for item.
     * 
     * We have a constraint (this) associated with item.  While the item 
     * does not yet have stop probabilities added to it, and so the current
     * constraint may or may not match it, we can check to make sure that
     * all constraints that are children of the current constraint do match
     * some child item of the given item.
     * 
     * It calls getAllConstraintsWithinItem to get an ArrayList of the
     * constraints to check, and then calls checkChildrenConstraints
     * to check that each constraint matches some item within the given item.
     * 
     * <b>Implementation note</b>: It is possible to improve the time 
     * efficiency of this function
     * by only checking for constraints that are now relevant because
     * of the item joined.  For example, if the modificand is
     * [1,5] and the modifier is [6,8], then there is no need to
     * check a constraint at [1,3] since it will already have been checked
     * before.  This would require passing in the modificand and
     * modifier, rather than the resulting item.
     */	 

    public boolean isViolatedByPartialConstituent(Item item) {
	CKYItem ckyItem = (CKYItem)item;

	if (baseIsNP((Symbol)item.label()) &&
	    !isLocallySatisfiedBy(item)) {
	    /*
	      System.out.println("eliminating specialNP case for " + 
	      " item = "+ item + 
	      " constraint = " + this);
	    */
	    return true;
	}
	
	boolean debug= false;
	ArrayList constraintsToCheck=
	    getAllConstraintsWithinItem(ckyItem.start(),ckyItem.end(),
					debug);

	checkChildrenConstraints(constraintsToCheck,ckyItem,true,debug);
	if (constraintsToCheck.size()==0) {
	    if (debug) {
		System.out.println("not violated by partial connstituent  " + 
				   this +  " onitem " + item);
	    }
	    return false;
	} else {
	    if (debug) {
		System.out.println("is violated by partial connstituent  " + 
				   this +  " onitem " + item);
		for (int jj=0; jj<constraintsToCheck.size(); jj++) {
		    System.out.println("  unfilled constraint "+constraintsToCheck.get(jj));
		}
	    }	
	    return true;
	}
    }

    /** 
     * @param itemStart start of the item's span
     * @param itemEnd end of the item's span.
     * @param debug  whether to print debugging info
     * @return an ArrayList of constraints that are descendents
     * of the current contraint that lie within
     * the span itemStart and itemEnd.  As soon as one such constraint is
     * found, searching along that branch of the constraint tree is stopped.
     * The idea is that if that constraint matches some item, then all its
     * children must be satisfied, by induction.
     */

    protected ArrayList getAllConstraintsWithinItem(int itemStart, 
						    int itemEnd, 
						    boolean debug) {
	ArrayList constraintsToCheck=new ArrayList();	
	if (debug) {
	    System.out.println("getAllConstraintsWithinItem " +
			       " itemStart="+itemStart+
			       " itemEnd="+itemEnd +
			       " this constraint = " + this);
	}
	getAllConstraintsWithinItemRecursive(itemStart,itemEnd,
					     this,constraintsToCheck,
					     debug,true);
	return constraintsToCheck;
    }

    /** 
     * @param itemStart start of the item's span
     * @param itemEnd end of the item's span.
     * @param constraint the purpose of this function is to find all the
     *        items that lie within this constraint.
     * @param constraintsToCheck contriants found get added to this ArrayList.
     * @param debug  whether to print debugging info
     * @param isRoot A horribly clunky flag indicating if we're at the top
     * of the constraint subtree that is being searched.  We only want the
     * descendent contraints, not the constraint currently associated with 
     * the item.
     */
    protected static void 
	getAllConstraintsWithinItemRecursive(int itemStart,
					     int itemEnd,
					     PartialTreeConstraint2 constraint,
					     ArrayList constraintsToCheck,
					     boolean debug,
					     boolean isRoot) {
	if (debug && !isRoot) {
	    System.out.println("   getAllConstraintsWithinItemRecursive "  +
			       " itemStart="+itemStart+
			       " itemEnd="+itemEnd +
			       " constraint = " + constraint);
	}
			       
			       
	if (!isRoot && constraintWithinItem(itemStart,itemEnd,
					    constraint.start(),
					    constraint.end())) {
	    if (debug) {
		System.out.println("    getAllConstraintsWithin adding " + 
				   constraint);
	    }
	    constraintsToCheck.add(constraint);
	    return;
	}
	Iterator it = constraint.getChildren().iterator();
	while (it.hasNext()) {
	    PartialTreeConstraint2 child = (PartialTreeConstraint2)it.next();
	
	    getAllConstraintsWithinItemRecursive(itemStart,itemEnd,
						 child,constraintsToCheck,
						 debug,false);
	}
    }


    /** 
     * @param itemStart start of the item's span
     * @param itemEnd end of the item's span.
     * @param contraintStart start of the constraint's span.
     * @param contraintEnd end of the constraint's span.
     * @return true if the constraint's span is within the item's span
     */
    protected static boolean constraintWithinItem(int itemStart, int itemEnd,
						  int constraintStart, 
						  int constraintEnd) {
	return (constraintStart>=itemStart && 
		constraintEnd<=itemEnd);
    }


    /**
     * Returns <code>true</code> if this constraint is satisfied by given
     * item.  The constraint is satisfied if the item is not inconsistent
     * with the constraint.  If the item matches the constraint, then
     * all child constraints match some item within the item's parse tree.
     * If the item does not match the constraint, then it does not match
     * negative contraint or is in any way inconsistent with the constraint.

     * @param item the item to test for satisfaction by this constraint
     * @return whether this constraint is satisfied the specified item
     *
     */

    public boolean isSatisfiedBy(Item item) {
	CKYItem ckyItem = (CKYItem)item;
	// normally, preterminal items should be assigned constraints from the
	// list of leaves, and thus this case should normally not be considered;
	// but just in case a brain-dead programmer creates a decoder that is
	// inefficient in this way, here's the code to deal with it
	// SK added:
	// that is, when an item is seeded in the chart, it is done so with stop=true
	// so there is no reason why addStopProbs should ever be calling 
	// isSatisfiedBy for a preterminal
	// Actually, this wasn't quite true.  constraintSatisfying, for the 
	// preterminal case (the only relevant one for this constraint scheme,
	// did in fact call constraint.isSatisfiedBy(item).  But there was no
	// need to, since it would just return true anyway.  So I modified
	// constraintSatisfying to just return the constraint for the 
	// preterminal case without calling isSatisfiedBy.  
	// There probably woulnd't be any harm in having it call isSatisfiedBy
	// anyway and letting isSatisfiedByPretermianl check that the start/end
	// is as expected, but that's probably going a bit too far with the 
	// checking.

	if (ckyItem.isPreterminal()) {
	    System.out.println("UNEXPECTED isSatisfiedBy for Preterminal"+item);
	    return isSatisfiedByPreterminal(ckyItem);
	}
	/*
	  if (!isLocallySatisfiedBy(item))
	  return false;
	*/
	if (!spanOK(item)) {
	    System.out.println("UNEXPECTED isSatisfiedBy !spanOK " + 
			       "\n    "+item+
			       "\n    "+this);
	    return false;
	}
	//    boolean debug=(start==13);
	//    boolean debug=(ckyItem.start()==3);
	boolean debug= false;
	boolean matchesNegative=matchesNegativeConstraint(item);
	boolean subsumes=subsumes(item);
	boolean isWildCard=(this.nt.base==Constants.kleeneStarSym);
	if (!spanMatches(item) && subsumes && !isWildCard && matchesNegative) {
	    // case 1a
	    return false;
	}
	/* 
	   at this point know that it can't be 1b,2,3,4b which are 
	   taken care of by isLocallySatisfied.  Check for them anyway
	   out of paranoia.  

	   !spanMatches - check for matchNegative  
	         subsumes && !isWildCard 
		              (case 1a, just checked)
		 subsumes && isWildCard 
		              (case 1b, checked before)
                 !subsumes (case 2, checked before)
	   spanMatches && !subsumes - check for matchesNegative 
                              (case 3, checked before)
	   spanMatches && subsumes && isWildcard - 
	                      (case 4b, checked before)
	   spanMatches && subsumes - check all internal constraints
	   (which shouldn't have any violations either, since these are 
	   all checked by isViolatedByPartialConstituent.

	*/

	if (!spanMatches(item)) {
	    if (matchesNegative) {
		boolean isSpecialNPCase=
		    baseIsNP((Symbol)ckyItem.label()) &&
		    baseIsBaseNP((Symbol)ckyItem.headChild().label()) &&
		    (ckyItem.leftChildren()!=null || 
		     ckyItem.rightChildren()!=null);
		if (isSpecialNPCase) {
		    System.out.println("UNEXPECTED isSpecialNPCase 1b or 2 item="+item + " constraint = " + this);
		} else {
		    System.out.println("UNEXPECTED case 1b or 2 item="+item + " constraint = " + this);
		}
		return false;
	    } else {
		return true;
	    }
	}

	if (!subsumes && !isWildCard) {
	    /* don't really need !isWildCard, since it must be !isWildCard, unless
	       there are function tags on * constraint that cause it to fail.
	       Otherwise, if the constraint is a wildcard, then it must 
	       subsume item
	   
	    */
	    if (matchesNegative) {
		System.out.println("UNEXPECTED case 3");
		return false;
	    } else {
		return true;
	    }
	}
	if (!subsumes) {
	    System.out.println("UNEXPECTED: impossible !subsume" + 
			       " item = "+ item + 
			       " constraint = " + this);
	    return false;
	}

	// of course it must be subsumes at this point
	if (subsumes && isWildCard && matchesNegative) {
	    boolean isSpecialNPCase=
		baseIsNP((Symbol)ckyItem.label()) &&
		baseIsBaseNP((Symbol)ckyItem.headChild().label()) &&
		(ckyItem.leftChildren()!=null || 
		 ckyItem.rightChildren()!=null);
	    if (isSpecialNPCase) {
		System.out.println("UNEXPECTED isSpecialNPCase 4b item="+item + " constraint = " + this);
	    } else {
		System.out.println("UNEXPECTED case 4b item="+ item + " constraint = " + this);
	    }
	    return false;
	}
	/* so now it is subsumes && spanMatches and 
	   (4a) !isWildcard  
	   (4b) isWildcard && !matchesNegative
	   so better make sure all internal constraints are satisfied
	   They should have been checked by isViolatedByPartialConstituent.  
	   But for now keep this
	   in and put in a warning to find if any such cases exist where
	   there is a violation here.  
	*/
	if (debug)
	    System.out.println("checking if item " + item + 
			       " satisfies constraint " + this);
	ArrayList constraintsNotSatisfied=
	    allChildConstraintsAreFullySatisfied(item,debug);
	if (constraintsNotSatisfied.size()==0) {
	    if (debug) {
		System.out.println("allowing constraint " + this +  
				   " onitem " + item);
	    }
	    return true;
	} else {
	    //	if (debug) {
	    if (true) {
		System.out.println("UNEXPECTED : disallowing constraint " + this + 
				   " onitem " + item);
	    }	
	    return false;
	}
    }



    /**
     * Returns <code>true</code>.
     * @param item a preterminal derivation item
     * @return <code>true</code> under all circumstances
     */
    protected boolean isSatisfiedByPreterminal(CKYItem item) {
	return true;
    }

    /**
     * Returns <code>true</code> if this item matches one of the specified 
     * negative constraints. This is the case if
     * {@link #constraintSatisfiedBySpuriousChild(item)} is false and the 
     * item's label is subsumed by one of the negative constraints.
     * @param item the chart item to check.
     * @return <code>true</code> if the item matches one of the negative 
     * constraints
     */     

    protected boolean matchesNegativeConstraint(Item item) {
	if (negativeConstraints==null) 
	    return false;
	if (item==null) {
	    System.err.println("WARNING: item is null in "+
			       " matchesNegativeConstraint!");
	    return true;
	}

	try {
	    Language.treebank().parseNonterminal((Symbol)item.label(), 
						 otherNT2);
	}
	catch (Exception e) {
	    System.err.println("WARNING: exception " +e + 
			       "from parseNonterminal");
	    e.printStackTrace();
	    System.out.println("item="+item);
	    System.out.println("otherNT2="+otherNT2);
	    return true;
	}

	if (constraintSatisfiedBySpuriousChild((CKYItem)item)) {
	    return false;
	}
	for (int i=0; i<negativeConstraints.size(); i++) {
	    Nonterminal negativeConstraintNT=
		(Nonterminal)negativeConstraints.get(i);
	    if (negativeConstraintNT.subsumes(otherNT2)) {
		return true;
	    }
	}
	return false;
    }

    /**
     * Returns <code>true</code> if this item should not be checked
     * for one of the negative contraints since it has an only child
     * that will be deleted.  
     * See separate documentation for details.
     * @param item the chart item to check.
     * @return <code>true</code> if the item has an only child that
     * will be deleted.
     */     
       
	
    protected boolean constraintSatisfiedBySpuriousChild(CKYItem item) {
	// only for NP item
	if (!baseIsNP((Symbol)item.label())) {
	    return false;
	}
	// must have just one child 
	if (item.leftChildren()!=null)
	    return false;
	if (item.rightChildren()!=null)
	    return false;
	CKYItem headChild=(CKYItem)item.headChild();

	// child is NPB
	if (!baseIsBaseNP((Symbol)headChild.label())) {
	    return false;
	}

	PartialTreeConstraint2 headChildConstraint=
	    (PartialTreeConstraint2)(headChild.getConstraint());
	// matches constraint.
	if (!headChildConstraint.subsumes(headChild) ||
	    !headChildConstraint.spanMatches(headChild)) {
	    return false;
	}
	/*
	  System.out.println("constraintSatisfiedByChild\n" +
	  "      item="+item+"\n"+
	  "      headChild="+headChild+"\n"+
	  "      headChildConstraint="+
	  headChildConstraint);
	*/
	return true;
    }


    /**
     * @param item item to check
     * @param debug flag to display debugging info
     * @return an ArrayList of all child constraints of the current constraint
     * that are not satisfied within the tree rooted by item
     *
     * Assumes that the current constraint is the constraint for the given item
     * (There really should be something cleaner)
     * It builds up a list of all the children of this constraint, and then
     * calls checkChildrenConstraints to see which ones match an an item
     * in the parse tree rooted by item.
     */

	    
    private ArrayList allChildConstraintsAreFullySatisfied(Item item,
							   boolean debug) {
	ArrayList constraintsToCheck=new ArrayList();
	Iterator it = children.iterator();
	while (it.hasNext()) {
	    PartialTreeConstraint2 child = (PartialTreeConstraint2)it.next();
	    if (debug) 
		System.out.println("   constraintsToCheck.add"+child);
	    constraintsToCheck.add(child);
	}
	checkChildrenConstraints(constraintsToCheck,(CKYItem)item,true,debug);

    
	return constraintsToCheck;

    }
    /**
     * @param constraintsToCheck  An ArrayList of constraints to make sure
     * match some descendent item of item.  
     * @param item The item within which to check for items that match 
     * the constraints in constraintsToCheck.
     * @param isRoot if we're at the root of the subtree rooted by item
     * from the first call.
     * Travervses through the parse tree represented by the given item.
     * For each item within it, it checks if it matches the constraint 
     * associated with that item.  If it does, then if that contraint is in 
     * contraintsToCheck, it removes that constraint from constraintsToCheck.  
     * In other words, at the end of this traversal of the given item, 
     * constraintsToCheck will consist only of those constraints that do not
     * match some item within the given item.
     * I'm not sure why I care if isRoot is true.  I think it's just for
     * debugging, since we don't expect the constraint to match the root
     */

    private void checkChildrenConstraints(ArrayList constraintsToCheck, 
					  CKYItem item,boolean isRoot,
					  boolean debug) {
	if (constraintsToCheck.size()==0) 
	    return;
	PartialTreeConstraint2 itemConstraint=
	    (PartialTreeConstraint2)(item.getConstraint());
	// if one of the constraints to check is a preterminal, then
	// we don't care about the label.  See comments under getParent
	boolean match=itemConstraint.matches(item);
	int index=constraintsToCheck.indexOf(itemConstraint);
	if (match && index>-1) {
	    if (isRoot) {
		System.out.println("WARNING: checkChildConstraints " + item  + 
				   "constraint="+item.getConstraint());
	    }
	    if (debug)
		System.out.println("   checkChildrenConstraints constraint " + 
				   item.getConstraint()+
				   " satisfied by  " + item);
	    constraintsToCheck.remove(index);
	}
	if (item.isPreterminal()) {
	    return;
	}

	checkChildrenConstraints(constraintsToCheck,item.headChild(),
				 false,debug);
	SLNode children=item.leftChildren();
	while (children!=null) {
	    CKYItem child=(CKYItem)children.data();
	    checkChildrenConstraints(constraintsToCheck,child,false,debug);
	    children=children.next();
	}

	children=item.rightChildren();
	while (children!=null) {
	    CKYItem child=(CKYItem)children.data();
	    checkChildrenConstraints(constraintsToCheck,child,false,debug);
	    children=children.next();
	}
	return;
    }

    /**
     * Returns whether this constraint has been <i>partially satisfied</i>.
     * @return whether this constraint has been <i>partially satisfied</i>.
     *
     * @see #satisfied
     */
    /* obviously not right but so what.
     */
    public boolean hasBeenSatisfied() { return true; }

    /**
     * Returns whether the specified item's span does not exceed the bounds
     * of the span associated with this constraint.
     * @param item the chart item whose span is to be tested
     * @return whether the specified item's span does not exceed the bounds
     * of the span associated with this constraint.
     */
    public boolean isLocallySatisfiedBy(Item item) {
	/*
	  return spanOK(item);
	*/
	if (!spanOK(item)) {
	    System.out.println("UNEXPECTED isLocallySatisfied !spanOK " + 
			       "\n    "+item+
			       "\n    "+this);
	    return false;
	}
	boolean debug= false;
	boolean matchesNegative=matchesNegativeConstraint(item);
	boolean subsumes=subsumes(item);
	boolean isWildCard=(this.nt.base==Constants.kleeneStarSym);
	// don't really need to check for !isWildCard since it follows
	// if !subsumes
	// case 2 - !spanMatches && !subsumes && matchesNegative
	// case 3 - spanMatches && !subsumes  && matchesNegative
	if (!subsumes && !isWildCard && matchesNegative) {
	    return false;
	}
	// case 1b - !spanMatches && isWildCard && matchesNegatie
	// case 4b - spanMatches & isWildcard && matchesNegative
	if (subsumes && isWildCard && matchesNegative) {
	    return false;
	}
	return true;
    }

    /**
     * Returns whether the span of the specified item crosses the span.
     * of this constraint
     * @param item the item whose span is to be tested
     * @return whether the span of the specified item crosses the span
     */
    protected boolean spanOK(Item item) {
	CKYItem ckyItem = (CKYItem)item;
	return ckyItem.start() >= this.start && ckyItem.end() <= this.end();
    }

    /**
     * Returns whether the start and end indices of the specified chart item are
     * equal to the start and end indices, respectively, of the span associated
     * with this constraint.
     *
     * @param item the chart item whose span is to be compared to that associated
     *             with this constraint
     * @return whether the start and end indices of the specified chart item are
     *         equal to the start and end indices, respectively, of the span
     *         associated with this constraint.
     */
    protected boolean spanMatches(Item item) {
	CKYItem ckyItem = (CKYItem)item;
	return ckyItem.start() == start && ckyItem.end() == end;
    }

    protected boolean subsumes(Item item) {
	Language.treebank().parseNonterminal((Symbol)item.label(), otherNT);
	return this.nt.subsumes(otherNT);
    }

    protected boolean matches(CKYItem item) {
	// check if item is preterminal, and if so only check span
	if (item.isPreterminal()) {
	    return spanMatches(item);
	} else {
	    return spanMatches(item) && subsumes(item);
	}
	//	return spanMatches(item) && subsumes(item);
    }





    /* 
       ==============================================================
       ASSORTED UTILITY FUNCTIONS
       ==============================================================
    */

  
    protected boolean baseIsNP(Symbol label) {
	return treebank.stripAugmentation(label)==NP;
    }
    protected boolean baseIsBaseNP(Symbol label) {
	return treebank.stripAugmentation(label)==baseNP;
    }
	      

    /**
     * Returns whether this constraint corresponds to a preterminal in the
     * original syntactic tree.
     * @return whether this constraint corresponds to a preterminal in the
     * original syntactic tree.
     */
    public boolean isLeaf() { return children.size() == 0; }

    /**
     * Returns the list of children of this constraint.
     * @return the list of children of this constraint.
     */
    protected List getChildren() { return children; }

    /**
     * Returns the nonterminal label associated with this constraint.
     * @return the nonterminal label associated with this constraint.
     */
    public Symbol label() { return label; }
    /**
     * Returns the start index of the span associated with this constraint.
     * @return the start index of the span associated with this constraint.
     */
    public int start() { return start; }
    /**
     * Returns the end index of the span associated with this constraint.
     * @return the end index of the span associated with this constraint.
     */
    public int end() { return end; }

    /**
     * Throws an {@link UnsupportedOperationException}, as this operation
     * is not appropriate for partial tree constraints.
     */
    public boolean isViolatedBy(Item item) {
	throw new UnsupportedOperationException();
    }



    /**
     * Returns a symbolic expression representing the tree of constraints rooted
     * at this constraint (intended for debugging purposes).
     *
     * @return a symbolic expression representing the tree of constraints rooted
     *         at this constraint
     */
    public Sexp toSexp() {
	SexpList retVal = new SexpList(children.size() + 1);

	Symbol thisNode = Symbol.add(label.toString() + "-" + start + "-" + end);

	retVal.add(thisNode);
	for (int i = 0; i < children.size(); i++)
	    retVal.add(((SexpConvertible)children.get(i)).toSexp());

	return retVal;
    }

    /**
     * Returns a human-readable string representation of this constraint (intended
     * for debugging purposes).
     * @return a human-readable string representation of this constraint.
     */
    public String toString() {
	return "label=" + label + ", span=(" + start + "," + end +
	    "), parentLabel=" +
	    (parent == null ? "null" : parent.label.toString());

    }

    public String toString(Item item) {
	return toString() +
	    " spanMatches="+spanMatches(item)+
	    " subsumes="+subsumes(item);
    }
}

      
